import inspect
import json
import os
import subprocess
import time
from pathlib import Path
from typing import Optional, Union

from pyslk.pyslk_exceptions import PySlkException

from ..base import ls
from ..constants import MAX_RETRIES_RETRIEVAL_ON_TIMEOUT, PYSLK_LOGGER, PYSLK_WILDCARDS
from ..core import _group_retrieve
from ..raw import retrieve2_raw, retrieve_raw

__all__ = [
    "retrieve",
    "retrieve_improved",
]


def retrieve_improved(
    resources: Union[
        Path, str, int, list[Path], list[str], list[int], set[Path], set[str], set[int]
    ],
    destination: Union[Path, str],
    dry_run: bool = False,
    force_overwrite: bool = False,
    ignore_existing: bool = False,
    resource_ids: bool = False,
    search_id: bool = False,
    recursive: bool = False,
    stop_on_failed_retrieval: bool = False,
    preserve_path: bool = True,
    verbose: bool = False,
) -> Optional[dict]:
    """Run a slk_helpers retrieve

    :param resources: multiple resource ids (set resource_ids=True) or paths OR one search id (set search_id=True)
    :type resources: str or Path or int or list[str] or list[Path] or list[int] or set[str] or set[Path] or set[int]
    :param destination: destination directory to check whether some of the needed files are already available and
        don't need to be retrieved
    :type destination: str or Path
    :param dry_run: list which file would be transferred but do not actually transfer data
    :type dry_run: bool
    :param force_overwrite: force overwrite of all existing files
    :type force_overwrite: bool
    :param ignore_existing: ignore-existing
    :type ignore_existing: bool
    :param resource_ids: consider input as 'resource' as resource id or resource ids
    :type resource_ids: bool
    :param search_id: consider input as 'resource' as search id
    :type search_id: bool
    :param recursive: use the -R flag to recall recursively, Default: False
    :type recursive: bool
    :param stop_on_failed_retrieval: stop immediately when one file cannot be retrieved
    :type stop_on_failed_retrieval: bool
    :param preserve_path: preserve original file path, Default: True
    :type preserve_path: bool
    :returns: dict of file paths which were or were not retrieved
    :rtype: Optional[dict]
    """
    destination = os.path.expandvars(os.path.expanduser(destination))
    # get json output of retrieval dry run; `slk_helpers retrieve -R --json-batch /arch/bm0146/k204221/iow -d .`
    output: subprocess.CompletedProcess = retrieve2_raw(
        resources=resources,
        destination=destination,
        dry_run=True,
        force_overwrite=force_overwrite,
        ignore_existing=ignore_existing,
        json_to_file=False,
        json_batch=True,
        print_progress=False,
        resource_ids=resource_ids,
        search_id=search_id,
        recursive=recursive,
        stop_on_failed_retrieval=stop_on_failed_retrieval,
        write_envisaged_to_file=False,
        write_missing_to_file=False,
        preserve_path=preserve_path,
        verbose=False,
        double_verbose=False,
        return_type=2,
    )
    """
    example json output:
    {
      "ENVISAGED": {
        "ENVISAGED": [
          "/arch/bm0146/k204221/iow/iow_data4_001.tar",
          "/arch/bm0146/k204221/iow/INDEX.txt"
        ]
      },
      "FAILED": {
        "FAILED_NOT_CACHED": [
          "/arch/bm0146/k204221/iow/iow_data_006.tar",
          "/arch/bm0146/k204221/iow/iow_data_005.tar",
          "/arch/bm0146/k204221/iow/iow_data_004.tar",
          "/arch/bm0146/k204221/iow/iow_data_003.tar",
          "/arch/bm0146/k204221/iow/iow_data_002.tar",
          "/arch/bm0146/k204221/iow/iow_data_001.tar"
        ]
      },
      "FILES": {
        "/arch/bm0146/k204221/iow/iow_data_006.tar": "/home/k204221/Documents/99_disk2tape/48_OperationalPhase/30_lost_files/42_accidental_cache_deletion_20240808/from_sl_db/./iow_data_006.tar",
        "/arch/bm0146/k204221/iow/iow_data_005.tar": "/home/k204221/Documents/99_disk2tape/48_OperationalPhase/30_lost_files/42_accidental_cache_deletion_20240808/from_sl_db/./iow_data_005.tar",
        "/arch/bm0146/k204221/iow/iow_data_004.tar": "/home/k204221/Documents/99_disk2tape/48_OperationalPhase/30_lost_files/42_accidental_cache_deletion_20240808/from_sl_db/./iow_data_004.tar",
        "/arch/bm0146/k204221/iow/iow_data_003.tar": "/home/k204221/Documents/99_disk2tape/48_OperationalPhase/30_lost_files/42_accidental_cache_deletion_20240808/from_sl_db/./iow_data_003.tar",
        "/arch/bm0146/k204221/iow/iow_data_002.tar": "/home/k204221/Documents/99_disk2tape/48_OperationalPhase/30_lost_files/42_accidental_cache_deletion_20240808/from_sl_db/./iow_data_002.tar",
        "/arch/bm0146/k204221/iow/iow_data_001.tar": "/home/k204221/Documents/99_disk2tape/48_OperationalPhase/30_lost_files/42_accidental_cache_deletion_20240808/from_sl_db/./iow_data_001.tar",
        "/arch/bm0146/k204221/iow/INDEX.txt": "/home/k204221/Documents/99_disk2tape/48_OperationalPhase/30_lost_files/42_accidental_cache_deletion_20240808/from_sl_db/./INDEX.txt"
      }
    }
    """
    # check if exit code larger than 2: directly throw error
    if output.returncode > 2:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )
    # check if we received proper JSON output
    try:
        output_json: dict = json.loads(output.stdout.decode("utf-8").rstrip())
    except json.decoder.JSONDecodeError:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + "parsing json output failed and slk_helpers yielded this error before: "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )
    # return JSON now, if dry_run was requested
    if dry_run:
        return output_json
    # if no envisaged files are present
    if (
        "ENVISAGED" not in output_json
        or "ENVISAGED" not in output_json["ENVISAGED"]
        or len(output_json["ENVISAGED"]["ENVISAGED"]) == 0
    ):
        return output_json
    # if there are any envisaged files
    retries: int
    for src_file_path in output_json["ENVISAGED"]["ENVISAGED"]:
        dst_file_path = os.path.expandvars(
            os.path.expanduser(output_json["FILES"][src_file_path])
        )
        print(f"attempting to retrieve {src_file_path} to {dst_file_path}")
        # initial exit code; should be 3 so that we enter the loop the first time
        retries = 0
        exit_code = 3
        accumulated_wait_time = 0
        while retries < MAX_RETRIES_RETRIEVAL_ON_TIMEOUT and exit_code == 3:
            # run retrieval
            output2: subprocess.CompletedProcess = retrieve2_raw(
                resources=src_file_path,
                destination=os.path.dirname(dst_file_path),
                dry_run=dry_run,
                force_overwrite=force_overwrite,
                ignore_existing=ignore_existing,
                json_to_file=False,
                json_batch=True,
                print_progress=False,
                resource_ids=False,
                search_id=False,
                recursive=False,
                stop_on_failed_retrieval=stop_on_failed_retrieval,
                write_envisaged_to_file=False,
                write_missing_to_file=False,
                preserve_path=False,
                verbose=False,
                double_verbose=False,
                return_type=2,
            )
            exit_code = output.returncode
            # if we have reached MAX_RETRIES_RETRIEVAL_ON_TIMEOUT, we will exit anyway => no need to wait further
            if exit_code == 3 and retries + 1 < MAX_RETRIES_RETRIEVAL_ON_TIMEOUT:
                retries += 1
                accumulated_wait_time += 10 * retries
                time.sleep(10 * retries)
        # check if exit code larger than 2: directly throw error
        if output.returncode == 3:
            raise PySlkException(
                f"pyslk.{inspect.stack()[0][3]}: retrieval of {src_file_path} failed due to repeated timeouts after "
                + f"{MAX_RETRIES_RETRIEVAL_ON_TIMEOUT} retries and a total wait time of {accumulated_wait_time} seconds"
            )
        # check if we received proper JSON output
        try:
            output2_json: dict = json.loads(output2.stdout.decode("utf-8").rstrip())
        except json.decoder.JSONDecodeError:
            raise PySlkException(
                f"pyslk.{inspect.stack()[0][3]}: "
                + f"parsing json output failed and 'slk_helpers retrieve {src_file_path}' yielded this error before: "
                + f"{output.stderr.decode('utf-8').rstrip()}"
            )
        # generally, remove path from envisaged section of output_json
        output_json["ENVISAGED"]["ENVISAGED"].remove(src_file_path)
        # add the path to the respective error or success section (there should be only one section!)
        key_lvl_1 = next(iter(output2_json))
        key_lvl_2 = next(iter(output2_json[key_lvl_1]))
        if key_lvl_1 not in output_json:
            output_json[key_lvl_1] = {}
        if key_lvl_2 not in output_json[key_lvl_1]:
            output_json[key_lvl_1][key_lvl_2] = []
        output_json[key_lvl_1][key_lvl_2].append(src_file_path)

    # all retrievals done return JSON
    return output_json


def retrieve(
    resource: Union[str, list, int],
    dest_dir: Union[str, None] = None,
    recursive: bool = False,
    group: Union[bool, None] = None,
    delayed: bool = False,
    **kwargs,
) -> Union[str, list]:
    """Retrieve data from tape archive using slk retrieve

    If group is True or resource is a list of files, the retrieve calls will
    be grouped by tape id to optimize retrieve calls.
    This functions aims at implementing the recommendations of DKRZ for
    `speeding up retrievals <https://docs.dkrz.de/doc/datastorage/hsm/retrievals.html#speed-up-your-retrievals>`_.

    :param resource:
        Path, pattern, SearchID, file or list of files that should be retrieved.
    :type resource: ``str``, ``list`` or ``int``
    :param dest_dir:
        Destination directory for retrieval. Retrieves to current directory by default.
    :type dest_dir: ``str``
    :param recursive:
        Retrieve recursively.
    :type recursive: ``bool``
    :param group:
        Group files by tape id and make one retrieve call per group. If group is None,
        retrieve calls will only be grouped if resource is a list of files.
        To totally avoid grouping, set group to False.
    :type group: ``bool``
    :param delayed:
        Delay retrieve calls using dask delayed.
    :type delayed: ``bool``
    :return: StdOut from slk calls. If delayed is True, a list of ``dask.delayed.Delayed``
        objects per tape_id is returned for later (maybe parallel) execution.
    :rtype: ``str`` or ``list``

    .. seealso::
        * :py:meth:`~pyslk.retrieve_raw`

    """

    if dest_dir is None:
        dest_dir = ""

    if group is True or (isinstance(resource, (list, int)) and group is None):
        return _group_retrieve(resource, dest_dir, recursive, delayed, **kwargs)

    if delayed is True:
        from dask import delayed as delay
    else:

        def delay(x):
            return x

    if isinstance(resource, str) and PYSLK_WILDCARDS in resource:
        # create file list, slk retrieve does not support this
        resource = list(ls(resource).filename)

    PYSLK_LOGGER.debug(f"resource: {resource}")

    if isinstance(resource, list):
        return [
            delay(retrieve_raw)(r, dest_dir, recursive=recursive, **kwargs)
            for r in resource
        ]

    return delay(retrieve_raw)(resource, dest_dir, recursive=recursive, **kwargs)


# def check_retrieve_permissions_dst(
#     path_or_id: Union[Path, str, int],
#     dest_dir: Union[str, Path],
#     recursive: bool = False,
#     duplicate: bool = False,
#     preserve_path: bool = False,
#     skip_if_exists: bool = False,
# ) -> set:
#     """Retrieve files via search id or GNS path.
#
#     Overwrite files if they already exists. Prevent this by 'duplicate'
#     or 'skip_if_exists'.
#
#     :param path_or_id: search id or gns path of resources to retrieve
#     :type path_or_id: ``str`` or ``int``
#     :param dest_dir: destination directory for retrieval
#     :type dest_dir: ``str``
#     :param recursive: use the -R flag to retrieve recursively, Default: False
#     :type recursive: ``bool``
#     :param duplicate: create a duplicate file if file exists
#     :type duplicate: ``bool``
#     :param preserve_path: preserve namespace in destination
#     :type preserve_path: ``bool``
#     :param skip_if_exists: Skip if file exists
#     :type skip_if_exists: ``bool``
#     :returns: stdout of the slk call
#     :rtype: ``str``
#     """
#     output = {"allowed": dict(), "not_allowed": dict(), "root_problem": dict()}
#     files = construct_target_file_path_local(path_or_id, dest_dir, recursive, preserve_path)



from os import path as op

import pytest

import pyslk as slk

from . import requires_slk, scratch

# once there is a test system ...
test_file = "test.txt"
arch_path = "/arch/ch0636/pyslk"


@pytest.fixture
def resource():
    with open(test_file, "w") as f:
        f.write("test content for pyslk")
    return test_file


@pytest.mark.new
@requires_slk
def test_with_slk(resource):
    # run slk commands once we have a test system
    out = slk.transfer.archive(resource=resource, dst_gns=arch_path)
    assert out == (["/arch/ch0636/pyslk/test.txt"], "Non-recursive Archive completed")

    path = op.join(arch_path, test_file)
    retr = slk.transfer.retrieve_improved(resources=path, destination=scratch)
    assert "SUCCESS" in retr
    assert "SUCCESS" in retr["SUCCESS"]
    assert path in retr["SUCCESS"]["SUCCESS"]
    assert "FILES" in retr
    assert path in retr["FILES"]


@requires_slk
def test_list():
    out = slk.list("/")
    assert "/arch" in out.filename.to_list()


@pytest.mark.new
@requires_slk
def test_retrieve():
    path = "/test/test3/ingest_01_1"
    # out = pyslk.transfer.retrieve.retrieve(path, scratch)
    out = slk.transfer.retrieve_improved(resources=path, destination=scratch)
    assert "SUCCESS" in out
    assert "SUCCESS" in out["SUCCESS"]
    assert path in out["SUCCESS"]["SUCCESS"]
    assert "FILES" in out
    assert path in out["FILES"]

    # retrieve
    path = ["/test/test3/ingest_01_1", "/test/test3/ingest_01_10"]
    # out = pyslk.transfer.retrieve.retrieve(path, scratch, group=False)
    out = slk.transfer.retrieve_improved(resources=path, destination=scratch)
    print(out)
    assert ("SUCCESS" in out) or ("SKIPPED" in out)
    assert ("SUCCESS" in out and "SUCCESS" in out["SUCCESS"]) or (
        "SKIPPED" in out and "SKIPPED_TARGET_EXISTS" in out["SKIPPED"]
    )
    assert ("SUCCESS" in out and path[0] in out["SUCCESS"]["SUCCESS"]) or (
        "SKIPPED" in out and path[0] in out["SKIPPED"]["SKIPPED_TARGET_EXISTS"]
    )
    # assert ('SUCCESS' in out and path[1] in out['SUCCESS']['SUCCESS']) or ('SKIPPED' in out and path[1] in out['SKIPPED']['SKIPPED_TARGET_EXISTS'])
    assert "FILES" in out
    assert path[0] in out["FILES"]
    # assert path[1] in out['FILES']

    # grouped retrieve
    # path = ["/test/test3/ingest_01_1", "/test/test3/ingest_01_10"]
    # out = slk.retrieve(path, scratch, group=True)
    # for o in out:
    #    assert o == ''



import re
from datetime import datetime

from ..raw import hostname_raw, session_raw
from ..utils import convert_expiration_date

__all__ = ["expiration_date", "session", "hostname", "valid_session", "valid_token"]


def valid_token() -> bool:
    """Returns whether session token is valid or not

    :returns: True if valid token exists; False otherwise
    :rtype: ``bool``
    """
    exit_code: int = session_raw(return_type=1)
    if exit_code == 0:
        return True
    else:
        return False


def valid_session() -> bool:
    """Returns whether session token is valid or not

    :returns: True if valid token exists; False otherwise
    :rtype: ``bool``
    """
    return valid_token()


def session() -> datetime:
    """Shows expiration date of your token

    :returns: expiration date of the login token
    :rtype: ``datetime``
    """
    output = session_raw()
    str_date = re.search(
        "[A-Z][a-z]{2} [A-Z][a-z]{2} [0-9]{2} [0-9]{2}:[0-9]{2}:[0-9]{2} [A-Z]+ [0-9]{4}",
        output,
    ).group(0)
    return convert_expiration_date(str_date)


def expiration_date() -> datetime:
    """Shows expiration date of your token

    :returns: expiration date of the login token
    :rtype: ``datetime``
    """
    return session()


def hostname() -> str:
    """Shows current hostname you are connected to

    :returns: hostname of the StrongLink system to which slk and slk_helpers currently connect to
    :rtype: ``str``
    """
    return hostname_raw()



import inspect
import subprocess
from typing import Union

from pyslk.constants import PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS, SLK, SLK_HELPERS
from pyslk.pyslk_exceptions import PySlkException
from pyslk.utils import login_valid, run_slk, which

__all__ = [
    "version_slk_raw",
    "version_slk_helpers_raw",
]


def version_slk_raw(
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """List the version of slk

    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    login_valid(rise_error_on_invalid=True)

    if which(SLK) is None:
        raise PySlkException(f"pyslk: {SLK}: command not found")

    slk_call = [SLK, "version"]

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def version_slk_helpers_raw(
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """List the version of slk_helpers

    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    slk_call = [SLK_HELPERS, "version"]

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )



import inspect
import os
import shutil
import subprocess
import typing
import warnings
from pathlib import Path
from typing import Union

from pyslk.constants import (
    PYSLK_DEFAULT_LIST_COLUMNS,
    PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
    SLK,
    SLK_HELPERS,
)
from pyslk.pyslk_exceptions import PySlkException
from pyslk.utils import FakeProc, _parse_list, run_slk, which

from .listing import list_raw

__all__ = [
    "recall_raw",
    "recall2_raw",
    "retrieve_raw",
    "retrieve2_raw",
    "archive_raw",
]


def recall_raw(
    path_or_id: Union[str, int, Path], recursive: bool = False, return_type: int = 0
) -> Union[str, int, subprocess.CompletedProcess, subprocess.Popen, FakeProc]:
    """Recall files from tape to cache via search id or GNS path

    :param path_or_id: search id or gns path of resources to recall
    :type path_or_id: str or int or Path
    :param recursive: use the -R flag to recall recursively, Default: False
    :type recursive: bool
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output), 3 (running subproc.)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess, subprocess.Popen, FakeProc]
    """
    if which(SLK) is None:
        raise PySlkException(f"pyslk: {SLK}: command not found")

    slk_call = [SLK, "recall"]
    if recursive:
        slk_call.append("-R")
    if isinstance(path_or_id, (str, Path)):
        slk_call.append(str(path_or_id))
    elif isinstance(path_or_id, int):
        slk_call.append(str(path_or_id))
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'path_or_id' needs to be 'str', 'int' or Path-like but got "
            + f"'{type(path_or_id).__name__}'."
        )

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    elif return_type == 3:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
            wait_until_finished=False,
        )
    elif return_type == -1:
        return FakeProc()
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def retrieve_raw(
    path_or_id: Union[str, int, Path],
    dest_dir: Union[str, Path],
    partsize: int = -1,
    streams: int = -1,
    stats: bool = False,
    writeblocksize: int = -1,
    recursive: bool = False,
    duplicate: bool = False,
    preserve_path: bool = True,
    skip_if_exists: bool = False,
    return_type: int = 0,
) -> typing.Optional[Union[str, int, subprocess.CompletedProcess, subprocess.Popen]]:
    """Retrieve files via search id or GNS path.

    Overwrite files if they already exists. Prevent this by 'duplicate'
    or 'skip_if_exists'.

    :param path_or_id: search id or gns path of resources to retrieve
    :type path_or_id: str or int or Path
    :param dest_dir: destination directory for retrieval
    :type dest_dir: str or Path
    :param partsize: size of each file to download per stream, Default: 500 (default is used if not >0)
    :type partsize: int
    :param streams: number of file part streams to use per node, Default: 4 (default is used if not >0)
    :type streams: int
    :param stats: Display part and I/O stats for each node
    :type stats: bool
    :param writeblocksize: I/O block size for file writes (KB), Default: 4096 (default is used if not >0)
    :type writeblocksize: int
    :param recursive: use the -R flag to retrieve recursively, Default: False
    :type recursive: bool
    :param duplicate: create a duplicate file if file exists
    :type duplicate: bool
    :param preserve_path: preserve namespace in destination [default: True]
    :type preserve_path: bool
    :param skip_if_exists: Skip if file exists
    :type skip_if_exists: bool
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output), 3 (running subproc.)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess, subprocess.Popen] or None

    .. seealso::
        * :py:meth:`~pyslk.retrieve`

    """
    if which(SLK) is None:
        raise PySlkException(f"pyslk: {SLK}: command not found")

    # check if a search id should be retrieved and if preserve_path is not set
    is_search: bool = False
    mv_files: bool = False
    src_paths: list = list()
    if isinstance(path_or_id, (str, int)):
        try:
            int(path_or_id)
            is_search = True
        except ValueError:
            pass
    if is_search and not preserve_path:
        warnings.warn(
            "Retrieving via search id without preserving namespace is not implemented in slk. "
            + "Instead slk will silently preserve the namespace. Pyslk will move the data after "
            + "successful download for you. However, this behaviour is not consistent with using "
            + "slk from the command line, hence, it will raise an error in the future or force "
            + "preserve_path=True. See https://gitlab.dkrz.de/hsm-tools/pyslk/-/issues/54."
        )
        mv_files = True
        # check for old slk version
        column_names = PYSLK_DEFAULT_LIST_COLUMNS
        try:
            list_out = list_raw(path_or_id)
        except PySlkException as e:
            if "ERROR: No resources found for given search id:" in str(e):
                warnings.warn("No resources found for given search id")
                return None
            else:
                raise e
        df = _parse_list(list_out, path_or_id, column_names, full_path=True)
        src_paths = df["filename"].tolist()
        # check for duplicate filenames
        tmp_files = [os.path.basename(src_path) for src_path in src_paths]
        duplicate_files = set()
        checked_file = set()
        for tmp_file in tmp_files:
            if tmp_file in checked_file:
                duplicate_files.add(tmp_file)
            else:
                checked_file.add(tmp_file)
        if len(duplicate_files) > 0:
            raise PySlkException(
                "You want to retrieve search results into one target directory without reconstructing the source path "
                + "('preserve_path' set to 'False' or not set at all). However, one or more search results have the "
                + "same name and would overwrite each other. Please run this command again with "
                + f"'preserve_path = True'. Affected filenames are: {', '.join(duplicate_files)}."
            )

    slk_call = [SLK, "retrieve"]
    if recursive:
        slk_call.append("-R")
    if isinstance(partsize, int):
        if partsize > 0:
            slk_call.append("--partsize")
            slk_call.append(str(partsize))
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'partsize' needs to be 'int' but got '{type(partsize).__name__}'."
        )
    if isinstance(streams, int):
        if streams > 0:
            slk_call.append("--streams")
            slk_call.append(str(streams))
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'streams' needs to be 'int' but got '{type(streams).__name__}'."
        )
    if stats:
        slk_call.append("--stats")
    if isinstance(writeblocksize, int):
        if writeblocksize > 0:
            slk_call.append("--writeblocksize")
            slk_call.append(str(writeblocksize))
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'writeblocksize' need to be 'int' but got "
            + f"'{type(writeblocksize).__name__}'."
        )
    if duplicate:
        slk_call.append("-d")
    if preserve_path:
        slk_call.append("-ns")
    if skip_if_exists:
        slk_call.append("-s")
    if isinstance(path_or_id, (str, int, Path)):
        slk_call.append(str(path_or_id))
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'path_or_id' need to be 'str', 'int' or path-like but got "
            + f"'{type(path_or_id).__name__}'."
        )
    if isinstance(dest_dir, (str, Path)):
        slk_call.append(str(dest_dir))
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'dest_dir' need to be 'str' or path-like but got "
            + f"'{type(dest_dir).__name__}'."
        )

    if return_type == 0:
        output = run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        output = run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        output = run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )

    if mv_files:
        for src_path in src_paths:
            shutil.move(
                os.path.join(dest_dir, src_path[1:]),
                os.path.join(dest_dir, os.path.basename(src_path)),
            )

    return output


def archive_raw(
    src_path: Union[str, list, Path],
    dst_gns: Union[str, Path],
    partsize: Union[int, None] = None,
    streams: Union[int, None] = None,
    stats: bool = False,
    recursive: bool = False,
    preserve_permissions: bool = False,
    exclude_hidden: bool = False,
    verbose: bool = False,
    double_verbose: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Upload files in a directory and optionally tag resources
            using directory path and GNS path

    :param src_path: search id or gns path of resources to retrieve
    :type src_path: str or list or Path
    :param dst_gns: destination directory for retrieval
    :type dst_gns: str or Path
    :param partsize: size of each file to download per stream, Default: 500
    :type partsize: int
    :param streams: number of file part streams to use per node, Default: 4
    :type streams: int
    :param stats: Display part & I/O stats for each node; Default: False
    :type stats: bool
    :param recursive: use the -R flag to archive recursively, Default: False
    :type recursive: bool
    :param preserve_permissions: preserve original file permission, Default: False
    :type preserve_permissions: bool
    :param exclude_hidden: exclude . (hidden) files, Default: False
    :type exclude_hidden: bool
    :param verbose: single verbose mode, Default: False
    :type verbose: bool
    :param double_verbose: double verbose mode, Default: False
    :type double_verbose: bool
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK) is None:
        raise PySlkException(f"pyslk: {SLK}: command not found")

    slk_call = [SLK, "archive"]
    if recursive:
        slk_call.append("-R")
    if preserve_permissions:
        slk_call.append("-p")
    if verbose:
        slk_call.append("-v")
    if double_verbose:
        slk_call.append("-vv")
    if exclude_hidden:
        slk_call.append("-x")
    if partsize is not None:
        if isinstance(partsize, int):
            slk_call.append("--partsize")
            slk_call.append(str(partsize))
        else:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'partsize' needs to be 'int' but got "
                + f"'{type(partsize).__name__}'"
            )
    if streams is not None:
        if isinstance(streams, int):
            slk_call.append("--streams")
            slk_call.append(str(streams))
        else:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'streams' needs to be 'int' but got "
                + f"'{type(streams).__name__}'"
            )
    if stats:
        slk_call.append("--stats")
    if isinstance(src_path, (str, Path)):
        slk_call.append(str(src_path))
    elif isinstance(src_path, list):
        if not all([isinstance(r, (str, Path)) for r in src_path]):
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: if argument 'src_path' is of type 'list' its items need to be"
                + "of type 'str' or Path-like but got type(s): "
                + f"'{', '.join([type(r).__name__ for r in src_path if not isinstance(r, (str, Path))])}'."
            )
        slk_call.extend(src_path)
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'src_path' needs to be 'str', 'list' or path-like but got "
            + f"'{type(src_path).__name__}'"
        )
    if isinstance(dst_gns, (str, Path)):
        slk_call.append(str(dst_gns))
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'dst_gns' needs to be 'str' or path-like but got "
            + f"'{type(dst_gns).__name__}'"
        )

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def recall2_raw(
    resources: Union[
        Path, str, int, list[Path], list[str], list[int], set[Path], set[str], set[int]
    ],
    destination: Union[Path, str, None] = None,
    resource_ids: bool = False,
    search_id: bool = False,
    recursive: bool = False,
    preserve_path: bool = True,
    verbose: bool = False,
    double_verbose: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Run a slk_helpers recall

    :param resources: multiple resource ids (set resource_ids=True) or paths OR one search id (set search_id=True)
    :type resources: str or Path or int or list[str] or list[Path] or list[int] or set[str] or set[Path] or set[int]
    :param destination: destination directory to check whether some of the needed files are already available and
        don't need to be recalled
    :type destination: str or Path
    :param resource_ids: consider input as 'resource' as resource id or resource ids
    :type resource_ids: bool
    :param search_id: consider input as 'resource' as search id
    :type search_id: bool
    :param recursive: use the -R flag to recall recursively, Default: False
    :type recursive: bool
    :param preserve_path: preserve original file path, Default: True
    :type preserve_path: bool
    :param verbose: single verbose mode, Default: False
    :type verbose: bool
    :param double_verbose: double verbose mode, Default: False
    :type double_verbose: bool
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    if resources:
        if isinstance(resources, (str, Path, int)):
            slk_call = [SLK_HELPERS, "recall", str(resources)]
        elif isinstance(resources, (list, set)):
            slk_call = [SLK_HELPERS, "recall"]
            slk_call.extend([str(r) for r in resources])
        else:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'resources' needs to be 'str', path-like, int or list/set "
                + f" containing variables of one of these types but got '{type(destination).__name__}'"
            )
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'resources' has to be set but got 'None'"
        )
    if destination:
        if isinstance(destination, (str, Path)):
            slk_call.append("-d")
            slk_call.append(str(destination))
        else:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'destination' needs to be 'str' or path-like but got "
                + f"'{type(destination).__name__}'"
            )
    if resource_ids:
        slk_call.append("--resource-ids")
    if search_id:
        slk_call.append("--search-id")
    if recursive:
        slk_call.append("-R")
    if preserve_path:
        slk_call.append("-ns")
    if verbose:
        slk_call.append("-v")
    if double_verbose:
        slk_call.append("-vv")

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def retrieve2_raw(
    resources: Union[
        Path, str, int, list[Path], list[str], list[int], set[Path], set[str], set[int]
    ],
    destination: Union[Path, str],
    dry_run: bool = False,
    force_overwrite: bool = False,
    ignore_existing: bool = False,
    json_to_file: Union[Path, str, None] = None,
    json_batch: bool = True,
    print_progress: bool = False,
    resource_ids: bool = False,
    search_id: bool = False,
    recursive: bool = False,
    stop_on_failed_retrieval: bool = False,
    write_envisaged_to_file: Union[Path, str, None] = None,
    write_missing_to_file: Union[Path, str, None] = None,
    preserve_path: bool = True,
    verbose: bool = False,
    double_verbose: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Run a slk_helpers retrieve

    :param resources: multiple resource ids (set resource_ids=True) or paths OR one search id (set search_id=True)
    :type resources: str or Path or int or list[str] or list[Path] or list[int] or set[str] or set[Path] or set[int]
    :param destination: destination directory to check whether some of the needed files are already available and
        don't need to be retrieved
    :type destination: str or Path
    :param dry_run: list which file would be transferred but do not actually transfer data
    :type dry_run: bool
    :param force_overwrite: force overwrite of all existing files
    :type force_overwrite: bool
    :param ignore_existing: ignore-existing
    :type ignore_existing: bool
    :param json_to_file: prints a JSON summary to a file (if you do not want to capture it together with verbose
        information)
    :type json_to_file: str or Path
    :param json_batch: prints a JSON summary to the terminal/stdout; all other output is suppressed or printed to stderr
    :type json_batch: bool
    :param print_progress: prints the progress of this command to stderr (if you do not want to capture it together with
        JSON or verbose information)
    :type print_progress: bool
    :param resource_ids: consider input as 'resource' as resource id or resource ids
    :type resource_ids: bool
    :param search_id: consider input as 'resource' as search id
    :type search_id: bool
    :param recursive: use the -R flag to recall recursively, Default: False
    :type recursive: bool
    :param stop_on_failed_retrieval: stop immediately when one file cannot be retrieved
    :type stop_on_failed_retrieval: bool
    :param write_envisaged_to_file:  write files, which could be retrieved, to provided file (failed files, which should
        be retrieved, are ignored)
    :type write_envisaged_to_file: str or Path
    :param write_missing_to_file: write files, which currently cannot / could not be retrieve, to provided file
    :type write_missing_to_file: str or Path
    :param preserve_path: preserve original file path, Default: True
    :type preserve_path: bool
    :param verbose: single verbose mode, Default: False
    :type verbose: bool
    :param double_verbose: double verbose mode, Default: False
    :type double_verbose: bool
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    if resources:
        if isinstance(resources, (str, Path, int)):
            slk_call = [SLK_HELPERS, "retrieve", str(resources)]
        elif isinstance(resources, (list, set)):
            slk_call = [SLK_HELPERS, "retrieve"]
            slk_call.extend([str(r) for r in resources])
        else:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'resources' needs to be 'str', path-like, int or list/set "
                + f" containing variables of one of these types but got '{type(destination).__name__}'"
            )
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'resources' has to be set but got 'None'"
        )

    if destination:
        if isinstance(destination, (str, Path)):
            slk_call.append("-d")
            slk_call.append(str(destination))
        else:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'destination' needs to be 'str' or path-like but got "
                + f"'{type(destination).__name__}'"
            )
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'destination' has to be set but got 'None'"
        )

    if dry_run:
        slk_call.append("--dry-run")
    if force_overwrite:
        slk_call.append("--force-overwrite")
    if ignore_existing:
        slk_call.append("--ignore-existing")
    if json_to_file:
        if isinstance(json_to_file, (str, Path)):
            slk_call.append("--json-to-file")
            slk_call.append(str(json_to_file))
        else:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'json_to_file' needs to be 'str' or path-like but got "
                + f"'{type(json_to_file).__name__}'"
            )
    if json_batch:
        slk_call.append("--json-batch")
    if print_progress:
        slk_call.append("--print-progress")
    if resource_ids:
        slk_call.append("--resource-ids")
    if search_id:
        slk_call.append("--search-id")
    if recursive:
        slk_call.append("-R")
    if stop_on_failed_retrieval:
        slk_call.append("--stop-on-failed-retrieval")
    if write_envisaged_to_file:
        if isinstance(write_envisaged_to_file, (str, Path)):
            slk_call.append("--write-envisaged-to-file")
            slk_call.append(str(write_envisaged_to_file))
        else:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'write_envisaged_to_file' needs to be 'str' or path-like "
                + f"but got '{type(write_envisaged_to_file).__name__}'"
            )
    if write_missing_to_file:
        if isinstance(write_missing_to_file, (str, Path)):
            slk_call.append("--write-missing-to-file")
            slk_call.append(str(write_missing_to_file))
        else:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'write_missing_to_file' needs to be 'str' or path-like "
                + f"but got '{type(write_missing_to_file).__name__}'"
            )
    if preserve_path:
        slk_call.append("-ns")
    if verbose:
        slk_call.append("-v")
    if double_verbose:
        slk_call.append("-vv")

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )



import inspect
import json
import re
import subprocess
import typing
from typing import Union

from ..pyslk_exceptions import PySlkException
from ..raw import (
    search_immediately_raw,
    search_incomplete_raw,
    search_raw,
    search_status_raw,
    search_successful_raw,
    searchid_exists_raw,
    total_number_search_results_raw,
)

__all__ = [
    "is_search_incomplete",
    "is_search_successful",
    "search",
    "search_immediately",
    "get_search_status",
    "searchid_exists",
    "total_number_search_results",
]


def search(search_string: str) -> int:
    """Performs a search based on a search_string can be provided.
    returns search id

    search_string has to be a valid JSON search string as described in the
    SLK-CLI manual. Simple double quotes have to be used in the JSON
    expression (no escaped double quotes, no escaped special characters).

    :param search_string: JSON search query string
    :type search_string: ``str``
    :returns: search id of the performed search
    :rtype: ``int``

    .. rubric:: Examples

    .. code-block:: python

        search_string='{"resources.mtime": {"$gt": "2021-09-02"}}'
    """
    if not isinstance(search_string, str):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'search_string' has wrong type; need 'str' but got "
            + f"{type(search_string).__name__}"
        )

    output: subprocess.CompletedProcess = search_raw(search_string, return_type=2)

    if output.returncode != 0:
        # an error
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )

    # no error /  get stdout
    stdout = output.stdout.decode("utf-8").rstrip()
    # if output has length 0 and does not contain a search id
    #  => we probably have an error in the search query
    if len(stdout) == 0:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: 'slk search' did not return any output. Probably, there is a tipo "
            + "in the JSON search query which is not properly captured by 'slk search'. Please verify your "
            + "JSON expression (e.g. via the command line tool 'jq'. You provided the search query "
            + f"'{search_string}'."
        )
    search_id_re = re.search("Search ID: [0-9]*", stdout)

    if search_id_re is not None:
        return int(search_id_re.group(0)[11:])

    raise PySlkException(
        f"pyslk.{inspect.stack()[0][3]}: search returned unexpected output: '{stdout}'. The search query was "
        + f"'{search_string}'."
    )


def search_immediately(search_string: str) -> int:
    """Performs a search based on a search_string and directly returns a search id while search is in progress

    search_string has to be a valid JSON search string as described in the
    SLK-CLI manual. Simple double quotes has to be used in the JSON
    expression (no escaped double quotes, no escaped special characters).

    :param search_string: JSON search query string
    :type search_string: ``str``
    :returns: search id of the performed search
    :rtype: ``int``

    .. rubric:: Examples

    .. code-block:: python

        search_string='{"resources.mtime": {"$gt": "2021-09-02"}}'
    """
    if not isinstance(search_string, str):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'search_string' has wrong type; need 'str' but got "
            + f"{type(search_string).__name__}"
        )

    output: subprocess.CompletedProcess = search_immediately_raw(
        search_string, return_type=2
    )

    if output.returncode != 0:
        # an error
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )

    # no error /  get stdout
    stdout = output.stdout.decode("utf-8").rstrip()
    # if output has length 0 and does not contain a search id
    #  => we probably have an error in the search query
    if len(stdout) == 0:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: 'slk search' did not return any output. Probably, there is a tipo "
            + "in the JSON search query which is not properly captured by 'slk search'. Please verify your "
            + "JSON expression (e.g. via the command line tool 'jq'. You provided the search query "
            + f"'{search_string}'."
        )

    try:
        return int(stdout)
    except ValueError:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: search returned unexpected output: '{stdout}'. The search query was "
            + f"'{search_string}'."
        )


def search_simple(fields_dict: dict) -> int:
    """Performs a search; key-value pairs from input dict are connected by AND;
    returns search id

    :param fields_dict: dict with "metadata-field: metadata-value" pairs
    :type fields_dict: ``dict``
    :returns: search id of the performed search
    :rtype: ``int``
    """
    # check if 'path' field is set
    if "path" in fields_dict:
        if not isinstance(fields_dict["path"], dict):
            if isinstance(fields_dict["path"], str):
                fields_dict["path"] = {"$gte": fields_dict["path"]}
            else:
                raise TypeError(
                    f"pyslk.{inspect.stack()[0][3]}: value of key 'path' has wrong type; has to be 'str' or 'dict' but "
                    + f"is of type '{type(fields_dict['path']).__name__}' with value '{fields_dict['path']}'."
                )
    # generate search string
    search_string = (
        '{"$and": ' + json.dumps([{k: v} for (k, v) in fields_dict.items()]) + "}"
    )
    return search(search_string)


def searchid_exists(search_id: Union[str, int]) -> bool:
    """Get path for a resource id

    :param search_id: a search_id
    :type search_id: ``str`` or ``int``
    :returns: whether provided search id exists (True) or not (False)
    :rtype: ``bool``
    """
    output: subprocess.CompletedProcess = searchid_exists_raw(search_id, return_type=2)
    if output.returncode == 0:
        return True
    elif output.returncode == 1:
        return False
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )


def is_search_incomplete(search_id: int) -> bool:
    """
    check if search is still running

    :param search_id: search id
    :type search_id: `int`
    :return: True if search is still running; False otherwise (might also be failed)
    :rtype: `bool`
    """
    output: subprocess.CompletedProcess = search_incomplete_raw(
        search_id, return_type=2
    )
    if output.returncode == 0:
        return True
    elif output.returncode == 1:
        return False
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )


def is_search_successful(search_id: int) -> bool:
    """
    check if search was successfully finished

    :param search_id: search id
    :type search_id: `int`
    :return: True if search finished successfully; False if still running; False if search failed
    :rtype: `bool`
    """
    output: subprocess.CompletedProcess = search_successful_raw(
        search_id, return_type=2
    )
    if output.returncode == 0:
        return True
    elif output.returncode == 1:
        return False
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )


def total_number_search_results(search_id: [int, str]) -> typing.Optional[int]:
    """Print the total number of search results regardless of read permissions of the current user

    :param search_id: a search_id
    :type search_id: ``str`` or ``int``
    :returns: total number of search results
    :rtype: ``int`` or ``None``
    """
    if not isinstance(search_id, (int, str)):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'search_id' has to be of type 'str' or 'int' but is "
            + f"{type(search_id).__name__}"
        )
    if isinstance(search_id, str):
        try:
            search_id = int(search_id)
        except ValueError:
            raise ValueError(
                f"pyslk.{inspect.stack()[0][3]}: cannot convert 'search_id' from 'str' to 'int'"
            )
    # search id does  not exist
    if not searchid_exists(search_id):
        return None
    # run the actual wrapper function
    output: str = total_number_search_results_raw(search_id, return_type=0)
    # return 'None' if output is 'None'
    if output is None:
        return None
    # else convert output to int and return it
    return int(output)


def get_search_status(search_id: int) -> dict:
    """
    return status of a search as dictionary

    :param search_id: search id
    :type search_id: `int`
    :return: dictionary which contains the search status and, if failed, a detailed error message
    :rtype: `dict`
    """
    output: subprocess.CompletedProcess = search_status_raw(
        search_id, json=True, return_type=2
    )
    if output.returncode in [0, 1]:
        return json.loads(output.stdout.decode("utf-8").rstrip())
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )



import inspect
import subprocess
from typing import Union

from pyslk.constants import PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS, SLK_HELPERS
from pyslk.pyslk_exceptions import PySlkException
from pyslk.utils import run_slk, which

__all__ = [
    "session_raw",
    "hostname_raw",
]


def session_raw(return_type: int = 0) -> Union[str, int, subprocess.CompletedProcess]:
    """Shows expiration date of your token

    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    slk_call = [SLK_HELPERS, "session"]

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def hostname_raw(return_type: int = 0) -> Union[str, int, subprocess.CompletedProcess]:
    """Shows current hostname you are connected to

    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    slk_call = [SLK_HELPERS, "hostname"]

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )



import inspect
import json
import os
import subprocess
import tempfile
import typing
from pathlib import Path
from typing import Union

from ..base import searchid_exists
from ..pyslk_exceptions import PySlkException
from ..raw import hsm2json_raw, json2hsm_raw, metadata_raw, tag_raw
from .resource_extras import resource_exists

__all__ = [
    "json2hsm",
    "json_file2hsm",
    "json_str2hsm",
    "json_dict2hsm",
    "hsm2json",
    "hsm2json_file",
    "hsm2json_dict",
    "get_metadata",
    "get_tag",
    "set_tag",
]


def json2hsm(
    json_file: Union[str, None] = None,
    restart_file: Union[str, None] = None,
    schema: Union[str, list, None] = None,
    expect_json_lines: bool = False,
    verbose: bool = False,
    quiet: bool = False,
    ignore_non_existing_metadata_fields: bool = False,
    write_mode: Union[str, None] = None,
    instant_metadata_record_update: bool = False,
    use_res_id: bool = False,
    skip_bad_metadata_sets: bool = False,
    json_string: Union[str, None] = None,
) -> dict:
    """
    Read metadata from JSON file and write them to archived files into HSM. Use absolute paths from metadata records
    to identify target files.

    :param json_file: JSON input file containing metadata
    :type json_file: ``str``
    :param restart_file: set a restart file in which the processed metadata entries are listed
    :type restart_file: ``str`` = None,
    :param schema: import only metadata fields of listed schemata; if str: comma-separated list without spaces
    :type schema: ``str`` or ``list``
    :param expect_json_lines: read JSON-lines from file instead of JSON
    :type expect_json_lines: ``bool``
    :param verbose: verbose mode
    :type verbose: ``bool``
    :param quiet: quiet mode
    :type quiet: ``bool``
    :param ignore_non_existing_metadata_fields: do not throw an error if a metadata field is used, which does not exist
            in StrongLink
    :type ignore_non_existing_metadata_fields: ``bool``
    :param write_mode: select write mode for metadata: OVERWRITE, KEEP, ERROR, CLEAN
    :type write_mode: ``str``
    :param use_res_id: use resource_id instead of path to identify file
    :type use_res_id: ``bool``
    :param skip_bad_metadata_sets: skip damaged / incomplete metadata sets [default: throw error]
    :type skip_bad_metadata_sets: ``bool``
    :param instant_metadata_record_update: False (default): read metadata records of all files and import into
            StrongLink afterward; True: import each metadata record after it has been read
    :type instant_metadata_record_update: ``bool``
    :param json_string: provide a json string instead of a json file; incompatible with json_file
    :type json_string: ``str``
    :returns: metadata import summary (key 'header')
    :rtype: ``dict``

    .. seealso::
        * :py:meth:`~pyslk.json_dict2hsm`
        * :py:meth:`~pyslk.json_file2hsm`
        * :py:meth:`~pyslk.json_str2hsm`
        * :py:meth:`~pyslk.hsm2json`
        * :py:meth:`~pyslk.hsm2json_dict`
        * :py:meth:`~pyslk.hsm2json_file`
    """
    if json_file is None and json_string is None:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: either 'json_file' xor 'json_string' have to be set but none of both is"
        )
    if json_file is not None and json_string is not None:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: only 'json_file' xor 'json_string' must to be set but both are"
        )
    out = json2hsm_raw(
        json_file,
        restart_file,
        schema,
        expect_json_lines,
        verbose,
        quiet,
        ignore_non_existing_metadata_fields,
        write_mode,
        instant_metadata_record_update,
        use_res_id,
        skip_bad_metadata_sets,
        print_summary=False,
        print_json_summary=True,
    )
    return {"header": json.loads(out)["_summary"]}


def json_file2hsm(
    json_file: str,
    restart_file: Union[str, None] = None,
    schema: Union[str, list, None] = None,
    expect_json_lines: bool = False,
    verbose: bool = False,
    quiet: bool = False,
    ignore_non_existing_metadata_fields: bool = False,
    write_mode: Union[str, None] = None,
    instant_metadata_record_update: bool = False,
    use_res_id: bool = False,
    skip_bad_metadata_sets: bool = False,
) -> dict:
    """
    Read metadata from JSON file and write them to archived files into HSM. Use absolute paths from metadata records
    to identify target files.

    :param json_file: JSON input file containing metadata
    :type json_file: ``str``
    :param restart_file: set a restart file in which the processed metadata entries are listed
    :type restart_file: ``str`` = None,
    :param schema: import only metadata fields of listed schemata; if str: comma-separated list without spaces
    :type schema: ``str`` or ``list``
    :param expect_json_lines: read JSON-lines from file instead of JSON
    :type expect_json_lines: ``bool``
    :param verbose: verbose mode
    :type verbose: ``bool``
    :param quiet: quiet mode
    :type quiet: ``bool``
    :param ignore_non_existing_metadata_fields: do not throw an error if a metadata field is used, which does not exist
            in StrongLink
    :type ignore_non_existing_metadata_fields: ``bool``
    :param write_mode: select write mode for metadata: OVERWRITE, KEEP, ERROR, CLEAN
    :type write_mode: ``str``
    :param use_res_id: use resource_id instead of path to identify file
    :type use_res_id: ``bool``
    :param skip_bad_metadata_sets: skip damaged / incomplete metadata sets [default: throw error]
    :type skip_bad_metadata_sets: ``bool``
    :param instant_metadata_record_update: False (default): read metadata records of all files and import into
            StrongLink afterward; True: import each metadata record after it has been read
    :type instant_metadata_record_update: ``bool``
    :returns: metadata import summary (key 'header')
    :rtype: ``dict``

    .. seealso::
        * :py:meth:`~pyslk.json2hsm`
        * :py:meth:`~pyslk.json_dict2hsm`
        * :py:meth:`~pyslk.json_str2hsm`
        * :py:meth:`~pyslk.hsm2json`
        * :py:meth:`~pyslk.hsm2json_dict`
        * :py:meth:`~pyslk.hsm2json_file`
    """
    return json2hsm(
        json_file,
        restart_file,
        schema,
        expect_json_lines,
        verbose,
        quiet,
        ignore_non_existing_metadata_fields,
        write_mode,
        instant_metadata_record_update,
        use_res_id,
        skip_bad_metadata_sets,
        None,
    )


def json_str2hsm(
    json_string: str,
    restart_file: Union[str, None] = None,
    schema: Union[str, list, None] = None,
    expect_json_lines: bool = False,
    verbose: bool = False,
    quiet: bool = False,
    ignore_non_existing_metadata_fields: bool = False,
    write_mode: Union[str, None] = None,
    instant_metadata_record_update: bool = False,
    use_res_id: bool = False,
    skip_bad_metadata_sets: bool = False,
) -> dict:
    """
    Read metadata from JSON file and write them to archived files into HSM. Use absolute paths from metadata records
    to identify target files.

    :param json_string: JSON string containing metadata
    :type json_string: ``str``
    :param restart_file: set a restart file in which the processed metadata entries are listed
    :type restart_file: ``str`` = None,
    :param schema: import only metadata fields of listed schemata; if str: comma-separated list without spaces
    :type schema: ``str`` or ``list``
    :param expect_json_lines: read JSON-lines from file instead of JSON
    :type expect_json_lines: ``bool``
    :param verbose: verbose mode
    :type verbose: ``bool``
    :param quiet: quiet mode
    :type quiet: ``bool``
    :param ignore_non_existing_metadata_fields: do not throw an error if a metadata field is used, which does not exist
            in StrongLink
    :type ignore_non_existing_metadata_fields: ``bool``
    :param write_mode: select write mode for metadata: OVERWRITE, KEEP, ERROR, CLEAN
    :type write_mode: ``str``
    :param use_res_id: use resource_id instead of path to identify file
    :type use_res_id: ``bool``
    :param skip_bad_metadata_sets: skip damaged / incomplete metadata sets [default: throw error]
    :type skip_bad_metadata_sets: ``bool``
    :param instant_metadata_record_update: False (default): read metadata records of all files and import into
            StrongLink afterward; True: import each metadata record after it has been read
    :type instant_metadata_record_update: ``bool``
    :returns: metadata import summary (key 'header')
    :rtype: ``dict``

    .. seealso::
        * :py:meth:`~pyslk.json2hsm`
        * :py:meth:`~pyslk.json_dict2hsm`
        * :py:meth:`~pyslk.json_file2hsm`
        * :py:meth:`~pyslk.hsm2json`
        * :py:meth:`~pyslk.hsm2json_dict`
        * :py:meth:`~pyslk.hsm2json_file`
    """
    fd, tmp_json_file = tempfile.mkstemp(suffix=".json", prefix="metadata")
    with os.fdopen(fd, "w") as f:
        f.write(json_string)
    output = json_file2hsm(
        tmp_json_file,
        restart_file,
        schema,
        expect_json_lines,
        verbose,
        quiet,
        ignore_non_existing_metadata_fields,
        write_mode,
        instant_metadata_record_update,
        use_res_id,
        skip_bad_metadata_sets,
    )
    os.remove(tmp_json_file)
    return output


def json_dict2hsm(
    json_dict: dict,
    restart_file: Union[str, None] = None,
    schema: Union[str, list, None] = None,
    expect_json_lines: bool = False,
    verbose: bool = False,
    quiet: bool = False,
    ignore_non_existing_metadata_fields: bool = False,
    write_mode: Union[str, None] = None,
    instant_metadata_record_update: bool = False,
    use_res_id: bool = False,
    skip_bad_metadata_sets: bool = False,
) -> dict:
    """
    Read metadata from JSON dictionary and write them to archived files into HSM. Use absolute paths from metadata
    records to identify target files.

    :param json_dict: a dictionary representing JSON
    :type json_dict: ``dict``
    :param restart_file: set a restart file in which the processed metadata entries are listed
    :type restart_file: ``str`` = None,
    :param schema: import only metadata fields of listed schemata; if str: comma-separated list without spaces
    :type schema: ``str`` or ``list``
    :param expect_json_lines: read JSON-lines from file instead of JSON
    :type expect_json_lines: ``bool``
    :param verbose: verbose mode
    :type verbose: ``bool``
    :param quiet: quiet mode
    :type quiet: ``bool``
    :param ignore_non_existing_metadata_fields: do not throw an error if a metadata field is used, which does not exist
            in StrongLink
    :type ignore_non_existing_metadata_fields: ``bool``
    :param write_mode: select write mode for metadata: OVERWRITE, KEEP, ERROR, CLEAN
    :type write_mode: ``str``
    :param use_res_id: use resource_id instead of path to identify file
    :type use_res_id: ``bool``
    :param skip_bad_metadata_sets: skip damaged / incomplete metadata sets [default: throw error]
    :type skip_bad_metadata_sets: ``bool``
    :param instant_metadata_record_update: False (default): read metadata records of all files and import into
            StrongLink afterward; True: import each metadata record after it has been read
    :type instant_metadata_record_update: ``bool``
    :returns: metadata import summary (key 'header')
    :rtype: ``dict``

    .. seealso::
        * :py:meth:`~pyslk.json2hsm`
        * :py:meth:`~pyslk.json_file2hsm`
        * :py:meth:`~pyslk.json_str2hsm`
        * :py:meth:`~pyslk.hsm2json`
        * :py:meth:`~pyslk.hsm2json_dict`
        * :py:meth:`~pyslk.hsm2json_file`
    """
    return json2hsm(
        None,
        restart_file,
        schema,
        expect_json_lines,
        verbose,
        quiet,
        ignore_non_existing_metadata_fields,
        write_mode,
        instant_metadata_record_update,
        use_res_id,
        skip_bad_metadata_sets,
        json.dumps(json_dict),
    )


def hsm2json(
    resources: Union[str, Path, list, None] = None,
    search_id: int = -1,
    recursive: bool = False,
    outfile: Union[str, Path, None] = None,
    restart_file: Union[str, Path, None] = None,
    schema: Union[str, list, None] = None,
    write_json_lines: bool = False,
    write_mode: Union[str, None] = None,
    instant_metadata_record_output: bool = False,
    print_hidden: bool = False,
) -> dict[str, Union[dict, list, None]]:
    """
    Extract metadata from HSM file(s) and return them in JSON structure

    :param resources: list of resources to be searched for
    :type resources: ``str`` or ``Path`` or ``list``
    :param search_id: id of a search
    :type search_id: ``int``
    :param recursive: export metadata from all files in gns_path recursively
    :type recursive: ``bool``
    :param outfile: Write the output into a file instead to the stdout
    :type outfile: ``str`` or ``Path`` or ``None``
    :param restart_file: set a restart file in which the processed metadata entries are listed
    :type restart_file: ``str`` or ``Path`` or ``None``
    :param schema: import only metadata fields of listed schemata; if str: comma-separated list without spaces
    :type schema: ``str`` or ``list`` or ``None``
    :param write_json_lines: write JSON-lines instead of JSON
    :type write_json_lines: ``bool`` = False
    :param write_mode: applies when 'output' is set; possible values: OVERWRITE, ERROR
    :type write_mode: ``str`` = None
    :param instant_metadata_record_output: False (default): read metadata of all files and write/print out
            afterward; True: write/print each metadata record after it has been read (requires 'write_json_lines')
    :type instant_metadata_record_output: ``bool``
    :param print_hidden: print read-only not-searchable metadata fields (sidecar file) [default: False]
    :type print_hidden: ``bool``
    :returns: dictionary with keys 'header' (summary report), 'metadata' (actual metadata) and 'file' (JSON file);
        either 'metadata' or 'file' is none depending on the value of input argument 'outfile'
    :rtype: ``dict``

    .. seealso::
        * :py:meth:`~pyslk.hsm2json_dict`
        * :py:meth:`~pyslk.hsm2json_file`
        * :py:meth:`~pyslk.json2hsm`
        * :py:meth:`~pyslk.json_dict2hsm`
        * :py:meth:`~pyslk.json_file2hsm`
        * :py:meth:`~pyslk.json_str2hsm`
    """
    if resources is None and search_id == -1:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: either 'resources' xor 'search_id' have to be set but none of both is"
        )
    if resources is not None and search_id != -1:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: only 'resources' xor 'search_id' must to be set but both are"
        )
    if outfile is None and (instant_metadata_record_output or write_json_lines):
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: if 'outfile' is not set, 'instant_metadata_record_output' and "
            + "'write_json_lines' cannot be set"
        )
    if instant_metadata_record_output and not write_json_lines:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: 'instant_metadata_record_output' does only work, if 'write_json_lines' is "
            + "set"
        )
    if restart_file is not None and not instant_metadata_record_output:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: 'restart_file' can only be used when 'instant_metadata_record_output' is "
            + "set"
        )

    # check types
    if resources is not None:
        if isinstance(resources, list):
            if not all([isinstance(r, (str, Path)) for r in resources]):
                raise TypeError(
                    f"pyslk.{inspect.stack()[0][3]}: argument 'resources' has to be 'str' or path-like or 'list' of "
                    + "'str' or path-like but is a list which contains at least elements of these other types: "
                    + f"'{', '.join([type(r).__name__ for r in resources if not isinstance(r, (str, Path))])}'"
                )
        elif not isinstance(resources, (str, Path)):
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: 'resources' has wrong type; need 'str' or path-like or 'list' but "
                + f"got {type(outfile).__name__}"
            )
    if outfile is not None:
        if not isinstance(outfile, (str, Path)):
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: 'outfile' has wrong type; need 'str' or 'Path' but "
                + f"got {type(outfile).__name__}"
            )
        # check if output file exists
        if os.path.exists(outfile):
            raise FileExistsError(
                f"pyslk.{inspect.stack()[0][3]}: 'outfile' does already exist; please remove it or set 'outfile' to "
                + f"another value; current value of 'outfile': {str(outfile)}"
            )

    # check types
    if restart_file is not None:
        if not isinstance(restart_file, (str, Path)):
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: 'restart_file' has wrong type; need 'str' or 'Path' but "
                + f"got {type(restart_file).__name__}"
            )
    output = hsm2json_raw(
        resources,
        search_id,
        recursive,
        outfile,
        restart_file,
        schema,
        write_json_lines,
        False,
        False,
        write_mode,
        False,
        True,
        True,
        instant_metadata_record_output,
        print_hidden,
    )
    if outfile is None:
        split_output = output.split("\n")
        json_string = json.loads(split_output[0])
        summary = json.loads(split_output[1])["_summary"]
    else:
        json_string = None
        summary = json.loads(output)["_summary"]

    return {"header": summary, "metadata": json_string, "file": outfile}


def hsm2json_file(
    outfile: str,
    resources: Union[str, Path, list] = "",
    search_id: int = -1,
    recursive: bool = False,
    restart_file: Union[str, None] = None,
    schema: Union[str, list, None] = None,
    write_json_lines: bool = False,
    write_mode: Union[str, None] = None,
    instant_metadata_record_output: bool = False,
    print_hidden: bool = False,
) -> None:
    """
    Extract metadata from HSM file(s) and return them in JSON structure

    :param outfile: Write the output into a file instead to the stdout
    :type outfile: ``str`` or ``Path``
    :param resources: list of resources to be searched for
    :type resources: ``str`` or ``Path`` or ``list``
    :param search_id: id of a search
    :type search_id: ``int``
    :param recursive: export metadata from all files in gns_path recursively
    :type recursive: ``bool``
    :param restart_file: set a restart file in which the processed metadata entries are listed
    :type restart_file: ``str`` = None,
    :param schema: import only metadata fields of listed schemata; if str: comma-separated list without spaces
    :type schema: ``str``, list or None
    :param write_json_lines: write JSON-lines instead of JSON
    :type write_json_lines: ``bool`` = False
    :param write_mode: applies when 'output' is set; possible values: OVERWRITE, ERROR
    :type write_mode: ``str`` = None
    :param instant_metadata_record_output: False (default): read metadata of all files and write/print out afterward;
            True: write/print each metadata record after it has been read (requires 'write_json_lines')
    :type instant_metadata_record_output: ``bool``
    :param print_hidden: print read-only not-searchable metadata fields (sidecar file) [default: False]
    :type print_hidden: ``bool``
    :returns: nothing; throws an error if writing failed
    :rtype: None

    .. seealso::
        * :py:meth:`~pyslk.hsm2json`
        * :py:meth:`~pyslk.hsm2json_dict`
        * :py:meth:`~pyslk.json2hsm`
        * :py:meth:`~pyslk.json_dict2hsm`
        * :py:meth:`~pyslk.json_file2hsm`
        * :py:meth:`~pyslk.json_str2hsm`
    """
    # check types
    hsm2json(
        resources,
        search_id,
        recursive,
        outfile,
        restart_file,
        schema,
        write_json_lines,
        write_mode,
        instant_metadata_record_output,
        print_hidden,
    )
    # return path of the output file
    return None


def hsm2json_dict(
    resources: Union[str, list] = "",
    search_id: int = -1,
    recursive: bool = False,
    restart_file: Union[str, None] = None,
    schema: Union[str, list, None] = None,
    print_hidden: bool = False,
) -> dict[str, Union[dict, list, None]]:
    """
    Extract metadata from HSM file(s) and return them in JSON structure

    :param resources: list of resources to be searched for
    :type resources: ``str`` or ``list``
    :param search_id: id of a search
    :type search_id: ``int``
    :param recursive: export metadata from all files in gns_path recursively
    :type recursive: ``bool``
    :param restart_file: set a restart file in which the processed metadata entries are listed
    :type restart_file: ``str`` = None,
    :param schema: import only metadata fields of listed schemata; if str: comma-separated list without spaces
    :type schema: ``str``, list or None
    :param print_hidden: print read-only not-searchable metadata fields (sidecar file) [default: False]
    :type print_hidden: ``bool``
    :returns: dictionary with keys 'header' (summary report), 'metadata' (actual metadata) and 'file' (None)
    :rtype: ``dict``

    .. seealso::
        * :py:meth:`~pyslk.hsm2json`
        * :py:meth:`~pyslk.hsm2json_file`
        * :py:meth:`~pyslk.json2hsm`
        * :py:meth:`~pyslk.json_dict2hsm`
        * :py:meth:`~pyslk.json_file2hsm`
        * :py:meth:`~pyslk.json_str2hsm`
    """
    return hsm2json(
        resources,
        search_id,
        recursive,
        None,
        restart_file,
        schema,
        False,
        None,
        False,
        print_hidden,
    )


def get_metadata(
    resource: Union[str, Path],
    print_hidden: bool = False,
    print_raw_values: bool = False,
) -> typing.Optional[dict[str, Union[str, int, float, dict]]]:
    """Get metadata

    :param resource: resource (full path)
    :type resource: ``str`` or ``Path``
    :param print_hidden: print read-only not-searchable metadata fields (sidecar file) [default: False]
    :type print_hidden: ``bool``
    :param print_raw_values: print metadata values without trying to convert them to int/float/dict [default: False]
    :type print_raw_values: ``bool``
    :returns: dictionary with the metadata
    :rtype: ``dict`` or ``None``

    .. seealso::
        * :py:meth:`~pyslk.get_tag`
        * :py:meth:`~pyslk.set_tag`
    """
    if not resource_exists(resource):
        return None
    output = metadata_raw(resource, True, print_hidden)
    out_dict = {
        o.split(": ")[0]: o.split(": ")[1] for o in output.split("\n") if o != ""
    }
    if not print_raw_values:
        # do some interpretations
        for k, v in out_dict.items():
            # decode JSON
            if v[0] == "{":
                # if we can decode the string as JSON expression, then we do this
                try:
                    out_dict[k] = json.loads(v)
                    continue
                except json.decoder.JSONDecodeError:
                    pass
            # try to convert string to int
            try:
                out_dict[k] = int(v)
                continue
            except ValueError:
                pass
            # try to convert string to float
            try:
                out_dict[k] = float(v)
                continue
            except ValueError:
                pass
    return out_dict


def get_tag(
    path_or_id: Union[str, int],
    recursive: bool = False,
) -> typing.Optional[dict]:
    """Apply metadata to the namespace and child resources

    :param path_or_id: search id or gns path of resources to retrieve
    :type path_or_id: ``str`` or ``int``
    :param recursive: use the -R flag to tag recursively, Default: False
    :type recursive: ``bool``
    :returns: metadata of the target files
    :rtype: ``dict`` or ``None``
    """
    # convert String search_id to int, if needed
    if isinstance(path_or_id, str):
        try:
            path_or_id = int(path_or_id)
        except ValueError:
            pass
    if isinstance(path_or_id, int):
        # if path_or_id is a search_id check whether the search ID exists
        if not searchid_exists(path_or_id):
            return None
    elif isinstance(path_or_id, (str, Path)):
        # if path_or_id is a str or Path check whether the resource exists
        if not resource_exists(path_or_id):
            return None
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'path_or_id' needs to be 'str', 'int' or Path-like but got "
            + f"'{type(path_or_id).__name__}'"
        )
    # run tag_raw
    output: subprocess.CompletedProcess = tag_raw(
        path_or_id, recursive=recursive, display=True, return_type=2
    )
    # check if error occurred
    if output.returncode == 0:
        # print new metadata
        return json.loads(output.stdout.decode("utf-8").rstrip())
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )


def set_tag(
    path_or_id: Union[str, int],
    metadata: dict,
    recursive: bool = False,
) -> typing.Optional[dict]:
    """Apply metadata to the namespace and child resources

    :param path_or_id: search id or gns path of resources to retrieve
    :type path_or_id: ``str`` or ``int``
    :param metadata: dict that holds as keys "[metadata schema].[field]" and
                     as values the metadata values
    :type metadata: ``dict``
    :param recursive: use the -R flag to tag recursively, Default: False
    :type recursive: ``bool``
    :returns: new metadata of the target files
    :rtype: ``dict`` or ``None``
    """
    # convert String search_id to int, if needed
    if isinstance(path_or_id, str):
        try:
            path_or_id = int(path_or_id)
        except ValueError:
            pass
    if isinstance(path_or_id, int):
        # if path_or_id is a search_id check whether the search ID exists
        if not searchid_exists(path_or_id):
            return None
    elif isinstance(path_or_id, (str, Path)):
        # if path_or_id is a str or Path check whether the resource exists
        if not resource_exists(path_or_id):
            return None
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'path_or_id' needs to be 'str', 'int' or Path-like but got "
            + f"'{type(path_or_id).__name__}'"
        )
    if isinstance(metadata, dict):
        if len(metadata) == 0:
            raise ValueError(
                f"slk.{inspect.stack()[0][3]}: argument 'metadata' is an empty dictionary; needs to have at "
                + "least one key-value pair"
            )
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'metadata' needs to be 'dict' but got "
            + f"'{type(metadata).__name__}'"
        )
    # run tag_raw
    output: subprocess.CompletedProcess = tag_raw(
        path_or_id, metadata, recursive, False, return_type=2
    )
    # check if error occurred
    if output.returncode == 0:
        # print new metadata
        return get_tag(path_or_id, recursive)
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )



import inspect
import subprocess
from pathlib import Path
from typing import Union

from pyslk.constants import PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS, SLK_HELPERS
from pyslk.pyslk_exceptions import PySlkException
from pyslk.utils import run_slk, which

__all__ = [
    "job_status_raw",
    "job_exists_raw",
    "job_queue_raw",
    "job_report_raw",
]


def job_status_raw(
    job_id: int, return_type: int = 0
) -> Union[str, int, subprocess.CompletedProcess]:
    """Check the status of a tape read job with the given ID

    Possible status values are AVAILABLE, BLOCKED and ERRORSTATE

    :param job_id: id of a recall job in StrongLink
    :type job_id: int
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    if not isinstance(job_id, int):
        raise TypeError(
            f'pyslk.{inspect.stack()[0][3]}: argument "job_id" needs to be "int" but got "{type(job_id).__name__}".'
        )

    slk_call = [SLK_HELPERS, "job_status", str(job_id)]

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def job_exists_raw(
    job_id: int, return_type: int = 0
) -> Union[str, int, subprocess.CompletedProcess]:
    """Check if tape read job exists

    :param job_id: id of a recall job in StrongLink
    :type job_id: int
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    if not isinstance(job_id, int):
        raise TypeError(
            f'pyslk.{inspect.stack()[0][3]}: argument "job_id" needs to be "int" but got "{type(job_id).__name__}".'
        )

    slk_call = [SLK_HELPERS, "job_exists", str(job_id)]

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def job_queue_raw(
    interpret: Union[str, None] = None, return_type: int = 0
) -> Union[str, int, subprocess.CompletedProcess]:
    """Prints status of the queue of tape read jobs

    :param interpret: interpret the queue length; possible values: DETAILS, NUMBER, TEXT, RAW, JSON
    :type interpret: str or None
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    slk_call = [SLK_HELPERS, "job_queue"]

    if interpret is not None:
        slk_call.append("--interpret")
        slk_call.append(interpret)

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def job_report_raw(
    job_id: Union[int, str],
    outfile: Union[Path, str, None] = None,
    overwrite: bool = False,
    return_incomplete_report: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """
    return job report

    :param job_id: a job id
    :type job_id: `str` or `int`
    :param outfile: optional output file if job report should be written into a file
    :type outfile: `str` or `Path`
    :param overwrite: overwrite destination output file if it already exists
    :type overwrite: `bool`
    :param return_incomplete_report: try to write out job report even if job is not finished yet
    :type return_incomplete_report: `bool`
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: `int`
    :returns: stdout of the slk call
    :rtype: `Union[str, int, subprocess.CompletedProcess]`
    """
    if which(SLK_HELPERS) is None:
        raise RuntimeError(f"pyslk: {SLK_HELPERS}: command not found")

    slk_call = [SLK_HELPERS, "job_report"]

    # job_id
    if isinstance(job_id, int):
        slk_call.append(str(job_id))
    elif isinstance(job_id, str):
        try:
            slk_call.append(str(int(job_id)))
        except ValueError:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: 'job_id' cannot be processed; need 'str', which holds an integer "
                + f"number, or 'int' but got '{type(job_id).__name__}' with value '{job_id}'"
            )
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'job_id' has wrong type; need 'str' or 'int' but got "
            + f"'{type(job_id).__name__}'"
        )

    # --outfile
    if outfile is not None:
        if isinstance(outfile, (str, Path)):
            slk_call.append("--outfile").append(str(outfile))
        else:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: 'outfile' has wrong type; need 'str' or 'Path' but got "
                + f"'{type(outfile).__name__}'"
            )

    # --force-overwrite
    if not isinstance(overwrite, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'overwrite' has wrong type; need 'bool' but got "
            + f"'{type(overwrite).__name__}'"
        )
    if overwrite:
        slk_call.append("--force-overwrite")

    # --return-incomplete-report
    if not isinstance(return_incomplete_report, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'return_incomplete_report' has wrong type; need 'bool' but got "
            + f"'{type(return_incomplete_report).__name__}'"
        )
    if return_incomplete_report:
        slk_call.append("--return-incomplete-report")

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    elif return_type == 3:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
            wait_until_finished=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1, 2 or 3."
        )
